;; Copyright (c) 2006 Carlos Ungil

;; Permission is hereby granted, free of charge, to any person obtaining
;; a copy of this software and associated documentation files (the
;; "Software"), to deal in the Software without restriction, including
;; without limitation the rights to use, copy, modify, merge, publish,
;; distribute, sublicense, and/or sell copies of the Software, and to
;; permit persons to whom the Software is furnished to do so, subject to
;; the following conditions:

;; The above copyright notice and this permission notice shall be
;; included in all copies or substantial portions of the Software.

;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
;; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
;; LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
;; OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
;; WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(in-package :rcl)

#-cmu
(eval-when (:compile-toplevel :load-toplevel :execute)
  (let ((variable-name #+windows "PATH" #-windows "LD_LIBRARY_PATH")
        (separator #+windows ";" #-windows ":"))
    (let ((current #+sbcl (sb-posix:getenv variable-name)                    
		   #+ecl (ext:getenv variable-name)
		   #+openmcl (ccl::getenv variable-name)                   
                   #+allegro (sys:getenv variable-name)
		   #+clisp (ext:getenv variable-name)
                   #+lispworks (lispworks:environment-variable variable-name)))
      (unless (search *r-lib-path* current)
        (let ((new (concatenate 'string current
                                      (when current separator) *r-lib-path*)))
          #+sbcl (sb-posix:putenv (concatenate 'string variable-name "=" new))
          #+ecl (ext:setenv variable-name new)
          #+openmcl (ccl:setenv variable-name new)
          #+allegro (setf (sys:getenv variable-name) new)
          #+clisp (setf (ext:getenv variable-name) new)
          #+lispworks (setf (lispworks:environment-variable variable-name) new))))))
          

(defun set-r-home (r-home)
  #+sbcl (sb-posix:putenv (concatenate 'string "R_HOME=" r-home))
  #+ecl (ext:setenv "R_HOME" r-home)
  #+openmcl (ccl::setenv "R_HOME" r-home)
  #+allegro (setf (sys:getenv "R_HOME") r-home)
  #+clisp (setf (ext:getenv "R_HOME") r-home)
  #+lispworks (setf (lispworks:environment-variable "R_HOME") r-home)
  #+cmu (unless (equal (cdr (assoc :r_home ext:*environment-list*)) r-home)
	  (error "for cmucl the R_HOME variable has to be set before lisp is started (got ~A, expected ~A)" (cdr (assoc :r_home ext:*environment-list*)) r-home)))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defvar *r-lib-loaded* nil)
  (unless *r-lib-loaded*
    (pushnew *r-lib-path* cffi:*foreign-library-directories*)
    (cffi:load-foreign-library *r-lib*)
    (setf *r-lib-loaded* t)))

(defvar *r-session* nil)

(cffi:defcvar "R_SignalHandlers" :unsigned-long)
(cffi:defcvar "R_CStackLimit" :unsigned-long)

(defun disable-stack-checking ()
  (setf *r-cstacklimit* (1- (expt 2 32))))

(defun disable-signal-handling ()
  (setf *r-signalhandlers* 0))

(defun check-stack () ;; taken from rif&rossini's rclg
  (format t "STACK: LIMIT ~A, HANDLERS ~A~%" 
	  *R-CSTACKLIMIT* *R-SIGNALHANDLERS*)
  (force-output t))

(defun r-init ()
  (if *r-session*
      (warn "R already running!")
      (progn
	(set-r-home *r-home*)
 	(disable-signal-handling)
	(cffi:with-foreign-object (argv :pointer 4)
	  (setf (cffi:mem-aref argv :pointer 0) (cffi:foreign-string-alloc "rcl"))
	  (setf (cffi:mem-aref argv :pointer 1) (cffi:foreign-string-alloc "--silent"))
	  (setf (cffi:mem-aref argv :pointer 2) (cffi:foreign-string-alloc "--vanilla"))
	  (setf (cffi:mem-aref argv :pointer 3) (cffi:foreign-string-alloc "--max-ppsize=50000"))
	  (rf-initembeddedr 4 argv))
	;;#+(and linux (or sbcl cmu))
	(disable-stack-checking)
	(setf *r-session* t))))

#+(or) ;;this kills lisp as well!
(defun r-quit ()
  (if *r-session*
      (r-funcall "q" "no") ;; see Rpostscript.c
      (warn "R was not running"))
  (setf *r-session* nil))
